#include "global.c"




/* funkce pro heuristiku podle pomeru cena/vaha */
void solve_h() {
    // pole pomeru cena / vaha
    float heuristic[ITEMS];
    int i;
    for(i=0; i<items; i++) heuristic[i]=((float)item_cost[i]/(float)item_weight[i]);
    
    // hledej nejvyhodnejsi dosud nepouzite itemy
    int mass=0;
    int konec=0;
    int cost=0;

    while(!konec) {
        float h=0;
        int j=-1;
        konec=1;

        // j <- index dalsiho dosud nepouziteho nejvyhodnejsiho prvku
        for(i=0; i<items; i++) if(!solution[i] && heuristic[i]>h && (item_weight[i]+mass)<m) {
            j=i;
            h=heuristic[i];
            konec=0;
        }
        
        // prvek byl skutecne nalezen?
        if(!konec) {
            solution[j]=1;
            mass+=item_weight[j];
            cost+=item_cost[j];
        }
    }
    
    best_cost=cost;
}




/*
   funkce pro reseni pomoci hrube sily
   vraci celkovou cenu itemu ktere budou teprve pridany
   m je zbyvajici kapacita batohu
   i je aktualni item k vyzkouseni
   cost je nakumulovana cena veci v batohu
*/
// konfigurace reseni
char item_present[ITEMS];

void solve_b(int m, int i, int cost) {
    //printf(" level=%i, m=%i, cost=%i\n", i, m, cost);

    // prekrocena kapacita batohu?
    if(m<0) return;

    // prozkoumany vsechny itemy?
    if(i==items) {
        if(!best_cost || cost>best_cost) {
            best_cost=cost;
            memcpy(solution, item_present, items * sizeof(char));
        }
        return;
    }

    // zkus i-ty item zaradit do batohu
    item_present[i]=1;
    solve_b(m-item_weight[i], i+1, cost+item_cost[i]);

    // zkus i-ty item NEzaradit do batohu
    item_present[i]=0;
    solve_b(m, i+1, cost);
}

// vyprazdni pole item_present, vynuluj cenu
void clear_b() {
    int i;
    for(i=0; i<ITEMS; i++) item_present[i]=0;
}




// spust...
int main(int argc, char* argv[]) {
    // pole spravnych vysledku podle brute-force pro sledovani chyby heuristiky
    int real_cost[MAX_INSTANCES];
    float max_chyba=0;
    float sum_chyba=0;
    
    // brute force
    g_open(argv[1], "out_b.txt");
    g_start_time();
    while(g_read()) {
        printf(".");
        clear_b();
        solve_b(m, 0, 0);
        real_cost[instances]=best_cost;
        g_write();
    }
    g_stop_time();
    printf("\n");
    printf("brute-force - celkem sekund: %6.2f, v prumeru: %6.2f\n", g_time_total(), g_time());
    g_close();
    
    // heuristic
    g_open(argv[1], "out_h.txt");
    g_start_time();
    while(g_read()) {
        solve_h();
        float chyba=((float)real_cost[instances]-(float)best_cost)/best_cost;
        if(chyba>max_chyba) max_chyba=chyba;
        sum_chyba+=chyba;
        g_write();
    }
    g_stop_time();
    printf("heuristic   - celkem sekund: %6.2f, v prumeru: %6.2f\n", g_time_total(), g_time());
    printf("            - prumerna chyba:  %6.2f%%\n", (sum_chyba*100)/instances);
    printf("            - maximalni chyba: %6.2f%%\n", max_chyba*100);
    g_close();
}
